//------ DON'T TOUCH THIS PART! ------
#include <bits/epa-phasespace.h>
#include <bits/epa-process.h>
#include <bits/epa-jetfunc.h>
#include <nlo++-module_add.h>

//----- used namespaces -----
using namespace nlo;
using namespace std;


//----- declaration of the user defined functons -----
void inputfunc(unsigned int&, unsigned int&, unsigned int&);
void psinput(phasespace_epa *, double&);
user_base_epa * userfunc();

typedef unsigned long int (*module_add_type)(bool, const list<basic_string<char> >&, const basic_string<char>&);
extern  module_add_type module_add;


//----- array of the symbols symbols -----
extern "C"{
  
  struct { 
	const char *name;
	void *address;
  } user_defined_functions[] = 
  {
	//   process index: epa for e+e- --> jets
	{"procindex", (void *) "epa"},
  
	//   input function 
	{"inputfunc", (void *) inputfunc},
  
	//   phase space input function 
	{"psinput", (void *) psinput},
  
	//   user defined functions
	{"userfunc",  (void *) userfunc},
  
	//   module to generate the readable result
	{"main_module_add", (void *) module_add},
  
	//  end of the list
	{0, 0}
  };
}
//------ END OF THE DO-NOT-TOUCH-PART ------

//----------- USER DEFINED PART STARTS HERE ------------

class UserEPA : public user1h_epa
{
public:
  void initfunc(unsigned int);
  void userfunc(const event_epa&, const amplitude_epa&);
};

//----- defines the module to sum up the results of the different runs -----
module_add_type module_add = main_module_add<basic_user_result<user1h_epa::distbook_type> >;


user_base_epa * userfunc() {
  return new UserEPA;
}

void inputfunc(unsigned int& nj, unsigned int& nu, unsigned int& nd)
{
  //  number of jets
  nj = 4U;
  
  //  number of the up and down type flavours
  nu = 2U;
  nd = 3U;
} 

void psinput(phasespace_epa *ps, double& s)
{
  //  total c.m. energy square
  s = 8315.068969;
  
  //   You can use your own phase generator. 
  //   Here we use the default.
  ps = 0;
} 


void UserEPA::initfunc(unsigned int)
{
  phys(1, "D parameter", 25,  0.0,  1.0);
}

double dparameter(const event_epa&);

void UserEPA::userfunc(const event_epa& p, const amplitude_epa& amp)
{
  double d = dparameter(p);
  physfilld(1, d, d*amp());
}


double dparameter(const event_epa& p)
{
  int n = p.upper();
  double ET = p[-1].T()+p[0].T();
  
  double xx = 0.0, yy = 0.0, zz = 0.0;
  double xy = 0.0, xz = 0.0, yz = 0.0;
  
  for(int a = 1; a <= n; a++) {
    xx += p[a].X()*p[a].X()/p[a].T();
    yy += p[a].Y()*p[a].Y()/p[a].T();
    zz += p[a].Z()*p[a].Z()/p[a].T();
    xy += p[a].X()*p[a].Y()/p[a].T();
    xz += p[a].X()*p[a].Z()/p[a].T();
    yz += p[a].Y()*p[a].Z()/p[a].T();
  }
  
  return 27.0*(xx*yy*zz+2.0*xy*xz*yz-xx*yz*yz-yy*xz*xz-zz*xy*xy)/(ET*ET*ET); 
}
