//------ DON'T TOUCH THIS PART! ------
#include <bits/dis-phasespace.h>
#include <bits/dis-process.h>
#include <bits/dis-jetfunc.h>
#include <nlo++-module_add.h>

//----- used namespaces -----
using namespace nlo;
using namespace std;


//----- declaration of the user defined functons -----
void inputfunc(unsigned int&, unsigned int&, unsigned int&);
void psinput(phasespace_dis *, double&, double&, double&, double&, double&, double&, double&, double&);
user_base_dis * userfunc();

typedef unsigned long int (*module_add_type)(bool, const list<basic_string<char> >&, const basic_string<char>&);
extern  module_add_type module_add;


//----- array of the symbols symbols -----
extern "C"{
  
  struct { 
	const char *name;
	void *address;
  } user_defined_functions[] = 
  {
	//   process index: hhc for e + p --> jets (DIS)
	{"procindex", (void *) "dis"},
  
	//   input function 
	{"inputfunc", (void *) inputfunc},
  
	//   phase space input function 
	{"psinput", (void *) psinput},
  
	//   user defined functions
	{"userfunc",  (void *) userfunc},
  
	//   module to generate the readable result
	{"main_module_add", (void *) module_add},
  
	//  end of the list
	{0, 0}
  };
}
//------ END OF THE DO-NOT-TOUCH-PART ------


//------ USER defined part starts here ------
#include <kT_clus.h>
#include "pdf-cteq61.h"


class UserDIS : public user1d_dis
{
public:
  //   init and user function
  void initfunc(unsigned int);
  void userfunc(const event_dis&, const amplitude_dis&);
  
private:
  //   pdf
  pdf_cteq61 pdf;
  
  // algorithms
  kT_clus_dis durham;
  
  // the jet structore in breit & lab. frame
  bounded_vector<lorentzvector<double> > pjb, pjl; 
  
  //  event in breit frame
  event_dis pbreit;
  
  // boost the jet momenta back to the laboratory frame
  void boost_back_to_lab(const event_dis&);

  //  scale studies
  class physscales
  {
  public:
    //   constructor
    physscales() {}
    physscales(const amplitude_dis& aa, double hsr, double hsf)
	  : hard_scale_renorm(hsr), hard_scale_fact(hsf), amp(&aa) {}
	
    //   set the input parameters
    void operator()(const amplitude_dis& aa, double hsr, double hsf) {
      amp = &aa; hard_scale_renorm = hsr; hard_scale_fact = hsf;
    }
	
    //   calculate the weight in the term of scales
    weight_dis renorm(const double& x) const {
	  return amp -> operator()(x*x*hard_scale_renorm, hard_scale_fact);
    }
	
    weight_dis fact(const double& x) const {
      return amp -> operator()(hard_scale_renorm, x*x*hard_scale_fact);
    }
	
    weight_dis renorm_and_fact(const double& x) const {
      return amp -> operator()(x*x*hard_scale_renorm, x*x*hard_scale_fact);
    }
    
  private:
    //   private members
    double hard_scale_renorm, hard_scale_fact;
    const amplitude_dis *amp;
  };
  
  //   helper classes
  physscales scalefunc;
  
  pointer_to_unary_const_member_function<physscales, const double&, weight_dis>
	renorm, fact, renorm_fact;
  
public:
	//   default constructor
	UserDIS()
    : renorm(scalefunc, &physscales::renorm), fact(scalefunc, &physscales::fact),
	renorm_fact(scalefunc, &physscales::renorm_and_fact) {}    
};


//----- defines the module to sum up the results of the different runs -----
module_add_type module_add =  main_module_add<basic_user_result<user1d_dis::distbook_type> >;


user_base_dis * userfunc() {
  return new UserDIS;
}


void inputfunc(unsigned int& nj, unsigned int& nu, unsigned int& nd)
{
  //  number of jets
  nj = 3U;
  
  //  number of the up and down type flavours
  nu = 2U;
  nd = 3U;
} 

void psinput(phasespace_dis *ps, double& el, double& eh, double& q2min, double& q2max, 
			 double& xmin, double& xmax, double& ymin, double& ymax)
{
  //   we use the default phase space generator
  ps = 0;
  
  el = 27.439;    // GeV
  eh = 820.0;     // GeV
  
  //  Q^2 cuts
  q2min = 5.0;     // GeV^2
  q2max = 5000.0;  // GeV^2 
  
  //   xB cuts
  xmin = 0.0;
  xmax = 1.0;
  
  //   y cuts
  ymin = 0.2;
  ymax = 0.6;
}


void UserDIS::initfunc(unsigned int njet)
{
  phys(11, "Durham renorm scale dep. fcut =1",        50, 0.3, 2.0);
  phys(12, "Durham fact scale dep. fcut =1",          50, 0.3, 2.0);
  phys(13, "Durham renorm & fact scale dep. fcut =1", 50, 0.3, 2.0);
}

extern"C" double xalfaem_(double *);

double xalpha_em(double mq2) {
  return xalfaem_(&mq2);
}

void UserDIS::userfunc(const event_dis& p, const amplitude_dis& amp)
{
  typedef lorentzvector<double> _Lv;
  unsigned int njet = amp.npar();
  
  //----- copy the momenta and boost to the breit frame ----
  pbreit = p; lab_to_breit(pbreit);
  
  //----- do the cluster analysis-----
  double Q2 = -((p[-1] - p[-2]).mag2());
  const bounded_vector<double>& f = durham(sqrt(Q2), pbreit);
  
  //----- jet shape variable fn / fn*dsigma/dfn / -----
  double fl = f[njet+1], fh = f[njet];
  
  //----- scale dependece -----
  if(!(fl < 0.2 && fh > 0.2)) return;
  
  //----- jet structure in breit frame -----
  pjb.resize(1, (int) njet);
  durham.reco(njet, pjb);
  
  //----- jet structure in laboratory frame -----
  pjl = pjb; boost_back_to_lab(p);
  
  //----- cuts in the laboratory & breit frame -----
  double pt, ptsum = 0.0, ptmin = 5.0;        //  transverze energy
  double eta, emin = -1.0, emax = 3.5;        //  pseudo-rapidity cut
  for(unsigned int i = 1; i <= njet; i++) {
    pt = pjb[i].perp(); eta = pjl[i].prapidity();
    if(pt < ptmin || eta < emin || eta > emax) return;
    ptsum += pt; 
  }
  
  //----- hard scales -----
  double Q2r, Q2f;
  Q2r = Q2f = ptsum*ptsum;
  
  //----- set the pdf function and the overall factors -----
  double alem = xalpha_em(Q2);
  amp.pdf_and_qcd_coupling(pdf, 389385.730*alem*alem);

  if(amp.integral() == amplitude_dis::lo) pdf.mode(pdf_cteq61::lo);
  else pdf.mode(pdf_cteq61::nlo);
  
    //----- set the hard scattering scales -----
  scalefunc(amp, Q2r, Q2f);
  
  //----- calculate the scale dependences -----
  physfill(11, renorm);
  physfill(12, fact);
  physfill(13, renorm_fact);
}

void UserDIS::boost_back_to_lab(const event_dis& p)
{ 
  double x = p[-1].T()/p[hadron(0)].T();
  double bz = (1.0 - x)/(1.0 + x);
  threevector<double> bVec = -((pbreit[-1] + pbreit[hadron(0)]).boostVector());
  lorentzvector<double> p0(pbreit[hadron(0)]);
  
  p0.boost(bVec);
  double phi = p0.phi(), theta = p0.theta();
  unsigned int njet = pjl.upper();
  
  for(unsigned int i = 1; i <= njet; i++) {
    pjl[i].boost(bVec);
    pjl[i].rotateZ(-phi);
    pjl[i].rotateY(-theta); 
    pjl[i].boost(0.0, 0.0, bz);
  }
}
