//------ DON'T TOUCH THIS PART! ------
#include <bits/dis-phasespace.h>
#include <bits/dis-process.h>
#include <bits/dis-jetfunc.h>
#include <nlo++-module_add.h>

//----- used namespaces -----
using namespace nlo;
using namespace std;


//----- declaration of the user defined functons -----
void inputfunc(unsigned int&, unsigned int&, unsigned int&);
void psinput(phasespace_dis *, double&, double&, double&, double&, double&, double&, double&, double&);
user_base_dis * userfunc();

typedef unsigned long int (*module_add_type)(bool, const list<basic_string<char> >&, const basic_string<char>&);
extern  module_add_type module_add;


//----- array of the symbols symbols -----
extern "C"{
  
  struct { 
	const char *name;
	void *address;
  } user_defined_functions[] = 
  {
	//   process index: hhc for e + p --> jets (DIS)
	{"procindex", (void *) "dis"},
  
	//   input function 
	{"inputfunc", (void *) inputfunc},
  
	//   phase space input function 
	{"psinput", (void *) psinput},
  
	//   user defined functions
	{"userfunc",  (void *) userfunc},
  
	//   module to generate the readable result
	{"main_module_add", (void *) module_add},
  
	//  end of the list
	{0, 0}
  };
}
//------ END OF THE DO-NOT-TOUCH-PART ------


//------ USER defined part starts here ------
#include "h1jet.h"
#include "pdf-cteq61.h"


class UserDIS : public user1h_dis
{
public:
  //   init and user function
  void initfunc(unsigned int);
  void userfunc(const event_dis&, const amplitude_dis&);
  
private:
  //   pdf
  pdf_cteq61 pdf;
  
  // algorithms
  h1jet jetclus;
  
  // the jet structore in breit & lab. frame
  bounded_vector<lorentzvector<double> > pjb, pjl; 
  
  //  event in breit frame
  event_dis pbreit;
  
  // boost the jet momenta back to the laboratory frame
  void boost_back_to_lab(const event_dis&);
};


//----- defines the module to sum up the results of the different runs -----
module_add_type module_add =  main_module_add<basic_user_result<user1h_dis::distbook_type> >;


user_base_dis * userfunc() {
  return new UserDIS;
}


void inputfunc(unsigned int& nj, unsigned int& nu, unsigned int& nd)
{
  //  number of jets
  nj = 3U;
  
  //  number of the up and down type flavours
  nu = 2U;
  nd = 3U;
} 

void psinput(phasespace_dis *ps, double& el, double& eh, double& q2min, double& q2max, 
			 double& xmin, double& xmax, double& ymin, double& ymax)
{
  //   we use the default phase space generator
  ps = 0;
  
  el = 27.439;    // GeV
  eh = 820.0;     // GeV
  
  //  Q^2 cuts
  q2min = 5.0;     // GeV^2
  q2max = 5000.0;  // GeV^2 
  
  //   xB cuts
  xmin = 0.0;
  xmax = 1.0;
  
  //   y cuts
  ymin = 0.2;
  ymax = 0.6;
}


void UserDIS::initfunc(unsigned int njet)
{
  const user1h_dis::point_type Q2[] = {
    {5.0, 7.5, 10.0},      {10.0, 12.5, 15.0},    {15.0, 20.0, 25.0},
    {25.0, 30.0, 35.0},    {35.0, 45.0, 55.0},    {55.0, 70.0, 85.0}, 
    {85.0, 100.0, 115.0},  {115.0, 132.5, 150.0}, {150.0, 185.0, 220.0},
    {220.0, 285.0, 350.0}, {350.0, 525.0, 700.0}, {700.0, 2350.0, 5000.0}
  };
  
  phys(1, "kT dsigma/dQ2 xmu = 1",   12, Q2, 0.25);
  phys(2, "kT dsigma/dQ2 xmu = 0.5", 12, Q2, 0.25);
  phys(3, "kT dsigma/dQ2 xmu = 2",   12, Q2, 0.25);
}

extern"C" double xalfaem_(double *);

double xalpha_em(double mq2) {
  return xalfaem_(&mq2);
}

#define __fill_hist(id, __x, __w)					\
switch(itype) {								\
 case amplitude_dis::real: sphysfilld(id, __x, sphysreal(__w)); break;	\
 case amplitude_dis::sub:  sphysfilld(id, __x, sphyssub(__w));  break;	\
 default: physfilld(id, __x, __w);					\
}



void UserDIS::userfunc(const event_dis& p, const amplitude_dis& amp)
{
  typedef lorentzvector<double> _Lv;
  unsigned int njet = amp.npar();
  
  //----- type of the contribution -----
  amplitude_dis::contrib_type itype = amp.contrib();
  
  //----- cut on the outgoung lepton (theta > 13*pi/15) -----
  //     double Q2 = -((p[-1] - p[-2]).mag2());
  //     if(Q2 < 100 && p[-2].theta() <= 2.7227136) return;
  
  //----- copy the momenta and boost to the breit frame ----
  pbreit = p; lab_to_breit(pbreit);
  
  //----- do the cluster analysis-----
  pjb = jetclus(pbreit);
  unsigned int nj = pjb.upper(); 
  if(nj < njet) return;   //  unable to resolve njet
  
  //----- jet structure in laboratory frame -----
  pjl = pjb; boost_back_to_lab(p);
  
  //----- cuts -----
  double pt_sum = 0.0, pt, ptmin = 5.0;  //  transverse energy cut
  double elab, emin = -1.0, emax = 2.5;  //  pseudo-rapidity cut
  lorentzvector<double> psum;
  unsigned int jetnum = 0;
  
  for(unsigned int i = 1; i <= nj; i++) {
    pt = pjb[i].perp(); elab = pjl[i].prapidity();
    if(pt > ptmin && emin < elab && elab < emax) {
      psum += pjb[i];
      pt_sum += pt;
      jetnum++;
    }
  }
  
  if(jetnum != njet) return;  //  unable to resolve njet
  
  //----- jet mass cuts -----
  double Mn = psum.mag();
  if(Mn < 25) return;
  
  //----- set the pdf function and the overall factors -----
  double Q2 = -((p[-1] - p[-2]).mag2());
  double alem = xalpha_em(Q2);
  amp.pdf_and_qcd_coupling(pdf, 389385730.0*alem*alem);
  
  if(amp.integral() == amplitude_dis::lo) pdf.mode(pdf_cteq61::lo);
  else pdf.mode(pdf_cteq61::nlo);
  
  //----- weight of the event -----
  double mu2 = pt_sum/((double) njet); mu2 *= mu2; 
  weight_dis wt;
  
  wt = amp(mu2, mu2);
  __fill_hist(1, Q2, wt);
  
  wt = amp(0.25*mu2, 0.25*mu2);
  __fill_hist(2, Q2, wt);
  
  wt = amp(4.0*mu2, 4.0*mu2);
  __fill_hist(3, Q2, wt);
}


void UserDIS::boost_back_to_lab(const event_dis& p)
{ 
  double x = p[-1].T()/p[hadron(0)].T();
  double bz = (1.0 - x)/(1.0 + x);
  threevector<double> bVec = -((pbreit[-1] + pbreit[hadron(0)]).boostVector());
  lorentzvector<double> p0(pbreit[hadron(0)]);
  
  p0.boost(bVec);
  double phi = p0.phi(), theta = p0.theta();
  unsigned int njet = pjl.upper();
  
  for(unsigned int i = 1; i <= njet; i++) {
    pjl[i].boost(bVec);
    pjl[i].rotateZ(-phi);
    pjl[i].rotateY(-theta); 
    pjl[i].boost(0.0, 0.0, bz);
  }
}
