#ifndef __pdf_cteq61_h__
#define __pdf_cteq61_h__ 1


#include <algorithm>
#include <bits/dis-process.h>
#include <bits/nlo-color.h>


//----- used namespaces -----
using namespace nlo;
using namespace std;

extern "C" double ctq6pdf_(int *, double *, double *);
extern "C" int setctq6_(int *);


class pdf_cteq61 : public pdf_and_coupling_dis
{
public:
  //   pdf mode 
  enum mode_type { nopdf = -1, nlo = 1, lo = 3, lo1 = 4};
  
  //   constructor
  explicit pdf_cteq61(int md = 1)
    : _M_mode(md) {}
  
  //   set mode
  void mode(int md) { _M_mode = md;}
  
  //   get mode
  int mode() const { return _M_mode;}
  
  //   strong coupling
  double alpha_qcd(unsigned int nf, double mr2) 
  {
	double b0  = B0(nf);
	double lml = std::log(std::sqrt(mr2)/(_M_mode == 4 ? 0.165 : 0.226)); 
	
	double res = b0*lml;
	if(_M_mode == 4) return 1.0/res;
	res += B1(nf)*std::log(2.0*lml)/b0; 
	return 1.0/res;
  }

  //   the parton distribution function
  void hadron(double x, double Q2, unsigned int, unsigned int, double *f) 
  {
    double mf = sqrt(Q2);
    this -> _M_init();
    for(int i=-6; i <= 6; i++) f[i] = ctq6pdf_(&i, &x, &mf);
    std::swap(f[1],f[2]); std::swap(f[-1],f[-2]);
  }
  
private:
	int _M_mode;
  
  //   initialize the pdf table
  void _M_init()
  {
    static int first = -1;
    if(first != _M_mode) {
      setctq6_(&_M_mode);
      first = _M_mode;
    }
  }
};



#endif

