//------ DON'T TOUCH THIS PART! ------
#include <bits/dis-phasespace.h>
#include <bits/dis-process.h>
#include <bits/dis-jetfunc.h>
#include <nlo++-module_add.h>

//----- used namespaces -----
using namespace nlo;
using namespace std;


//----- declaration of the user defined functons -----
void inputfunc(unsigned int&, unsigned int&, unsigned int&);
void psinput(phasespace_dis *, double&, double&, double&, double&, double&, double&, double&, double&);
user_base_dis * userfunc();

typedef unsigned long int (*module_add_type)(bool, const list<basic_string<char> >&, const basic_string<char>&);
extern  module_add_type module_add;


//----- array of the symbols symbols -----
extern "C"{
  
  struct { 
	const char *name;
	void *address;
  } user_defined_functions[] = 
  {
	//   process index: hhc for e + p --> jets (DIS)
	{"procindex", (void *) "dis"},
  
	//   input function 
	{"inputfunc", (void *) inputfunc},
  
	//   phase space input function 
	{"psinput", (void *) psinput},
  
	//   user defined functions
	{"userfunc",  (void *) userfunc},
  
	//   module to generate the readable result
	{"main_module_add", (void *) module_add},
  
	//  end of the list
	{0, 0}
  };
}
//------ END OF THE DO-NOT-TOUCH-PART ------

//------ USER ROUTINE ------
#include <kT_clus.h>
#include "pdf-cteq6.h"


class UserDIS : public basic_user_set<user1d_dis, user1h_dis>
{
public:
  //   init and user function
  void initfunc(unsigned int);
  void userfunc(const event_dis&, const amplitude_dis&);
  
private:
  //   pdf
  pdf_cteq6 pdf;

  // algorithms
  kT_clus_dis durham;
  
  // the jet structore in breit & lab. frame
  bounded_vector<lorentzvector<double> > pjb, pjl; 
  
  //  event in breit frame
  event_dis pbreit;
  
  // boost the jet momenta back to the laboratory frame
  void boost_back_to_lab(const event_dis&);
  
  //   calculate the thrust and the thrust axis
  static void thrust(const event_dis&, double&, threevector<double>&);
  //   calculate the thrust major and the thrust major axis
  static void thrustM(const event_dis&, double&, threevector<double>&);
  
public:
  UserDIS() : durham(true) {}
};


//----- defines the module to sum up the results of the different runs -----
module_add_type module_add =  main_module_add<basic_user_result<user1d_dis::distbook_type, user1h_dis::distbook_type> >;


user_base_dis * userfunc() {
  return new UserDIS;
}


void inputfunc(unsigned int& nj, unsigned int& nu, unsigned int& nd)
{
  //  number of jets
  nj = 3U;
  
  //  number of the up and down type flavours
  nu = 2U;
  nd = 3U;
} 

void psinput(phasespace_dis *ps, double& el, double& eh, double& q2min, double& q2max, 
			 double& xmin, double& xmax, double& ymin, double& ymax)
{
  //   we use the default phase space generator
  ps = 0;
  
  el = 27.439;    // GeV
  eh = 820.0;     // GeV
  
  //  Q^2 cuts
  q2min = 5.0;   // GeV^2
  q2max = 5000;  // GeV^2 
  
  //   xB cuts
  xmin = 0.0;
  xmax = 1.0;
  
  //   y cuts
  ymin = 0.2;
  ymax = 0.6;
}



void UserDIS::initfunc(unsigned int njet)
{
  const user1h_dis::point_type Q2[] = {
    {0.0, 2.5, 5.0},      {5.0, 7.5, 10.0},       {10.0, 12.5, 15.0},
    {15.0, 20.0, 25.0},   {25.0, 30.0, 35.0},     {35.0, 45.0, 55.0},
    {55.0, 70.0, 85.0},   {85.0, 100.0, 115.0},   {115.0, 132.5, 150.0},
    {150.0, 185.0, 220.0},{220.0, 285.0, 350.0},  {350.0, 525.0, 700.0},
    {700.0, 2350.0, 5000.0},{5000.0, 7500.0, 10000.0}
  };

  user1h_dis::phys(1, "Q2", 14, Q2,  0.25);
  user1h_dis::phys(2, "xB", 50, 0.0, 1.0);
  user1h_dis::phys(3, "y", 50, 0.0, 1.0);
  user1h_dis::phys(5, "dsigma/dKout, xmu = 1",   100, 0.0,  2.0);
  user1h_dis::phys(6, "dsigma/dKout, xmu = 0.5", 100, 0.0,  2.0);
  user1h_dis::phys(7, "dsigma/dKout, xmu = 2",   100, 0.0,  2.0);
  user1h_dis::phys(8, "dsigma/dlnKout, xmu = 1",   100, -12.0,  1.0);
  user1h_dis::phys(9, "dsigma/dlnKout, xmu = 0.5", 100, -12.0,  1.0);
  user1h_dis::phys(10, "dsigma/dlnKout, xmu = 2",  100, -12.0,  1.0);
  user1h_dis::phys(11, "kT dsigma/df3, xmu=1"  , 50, 0.0,  1.0);
  user1h_dis::phys(12, "kT dsigma/df3, xmu=0.5", 50, 0.0,  1.0);
  user1h_dis::phys(13, "kT dsigma/df3, xmu=2"  , 50, 0.0,  1.0);
  user1h_dis::phys(14, "kT dsigma/dlgf3, xmu=1"  , 40, -4.0,  0.0);
  user1h_dis::phys(15, "kT dsigma/dlgf3, xmu=0.5", 40, -4.0,  0.0);
  user1h_dis::phys(16, "kT dsigma/dlgf3, xmu=2"  , 40, -4.0,  0.0);
  
  user1d_dis::phys(17, "kT n-jet cross section, xmu=1"  , 40, -4.0,  0.0);
  user1d_dis::phys(18, "kT n-jet cross section, xmu=0.5", 40, -4.0,  0.0);
  user1d_dis::phys(19, "kT n-jet cross section, xmu=2"  , 40, -4.0,  0.0);
}

extern"C" double xalfaem_(double *);

double xalpha_em(double mq2) {
  return xalfaem_(&mq2);
}

#define __fill_hist(id, __x, __w)                                          \
switch(itype) {                                                            \
 case amplitude_dis::real: sphysfilld(id, __x, sphysreal(__w)); break;     \
 case amplitude_dis::sub:  sphysfilld(id, __x, sphyssub(__w));  break;     \
 default: user1h_dis::physfilld(id, __x, __w);                             \
}

void UserDIS::userfunc(const event_dis& p, const amplitude_dis& amp)
{
  typedef lorentzvector<double> _Lv;
  unsigned int njet = 3U;
  
  //----- type of the contribution -----
  amplitude_dis::contrib_type itype = amp.contrib();
  
  //----- copy the momenta and boost to the breit frame ----
  pbreit = p; lab_to_breit(pbreit);
  unsigned int np = pbreit.upper();

  //----- cut on the rapidities in the breit frame -----
  double eta, emin = -3.0;        //  pseudo-rapidity cut
  for(unsigned int i = 1; i <= np; i++) {
    eta = pbreit[i].prapidity(); //cout << eta << endl;
    if(eta < emin) return;
  }
  
  //----- compute DIS variables -----
  double Q2 = -((p[-1] - p[-2]).mag2());
  double Q = sqrt(Q2);
  double xb = Q/(2.0*pbreit[hadron(0)].T());
  double s = 4.0*p[-1].T()*p[hadron(0)].T();
  double y = Q2/(s*xb);
  //
  //----- do the cluster analysis-----
  const bounded_vector<double>& f = durham(Q, pbreit);
  
  //----- jet shape variable fn / fn*dsigma/dfn / -----
  double fl = f[njet+1], fh = f[njet];
  double lgfl = log10(fl), lgfh = log10(fh);
  
  //----- hard scales -----
  double mu2r, mu2f;
  mu2r = mu2f = Q2;
  
  //----- set the pdf function and the overall factors -----
  //----- cross section is given in nbarns -----
  double alem = xalpha_em(Q2);
  amp.pdf_and_qcd_coupling(pdf, 389379.323*alem*alem);

  //----- weight of the event -----
  weight_dis wtc = amp(mu2r, mu2f);
  weight_dis wtl = amp(0.25*mu2r, 0.25*mu2f);
  weight_dis wth = amp(4.0*mu2r, 4.0*mu2f);

  double T = 0.0;
  threevector<double> axis;

  //----- thrust in the current hemisphere -----
  //thrust(pbreit, T, axis);
  //double tau = 1.0-T;

  //----- thrust major -----
  thrustM(pbreit, T, axis);

  //----- basic DIS kinematic variables -----
  __fill_hist(1, Q2, wtc);
  __fill_hist(2, xb, wtc);
  __fill_hist(3, y, wtc);

  //----- Kout with cut on y2 -----
  double y2cut = 0.1;
  if(f[njet-1] > y2cut) {
  //----- the event plane -----
    threevector<double> phad = p[hadron(0)]; phad /= phad.mag();
    threevector<double> plane = cross(phad,axis); plane /= plane.mag();
    double Kout = 0.0;
    for(unsigned int i = 1; i <= np; i++) { Kout += abs(plane*pbreit[i]); }
    Kout /= Q; double lgKout = log(Kout);

  //----- dsigma/dKout
    __fill_hist(5, Kout, wtc);
    __fill_hist(6, Kout, wtl);
    __fill_hist(7, Kout, wth);

  //----- dsigma/dlgKout
    __fill_hist(8,  lgKout, wtc);
    __fill_hist(9,  lgKout, wtl);
    __fill_hist(10, lgKout, wth);
  }

  //----- dsigma/df3
  __fill_hist(11, fh, wtc);
  __fill_hist(12, fh, wtl);
  __fill_hist(13, fh, wth);

  //----- dsigma/dlgf3
  __fill_hist(14, lgfh, wtc);
  __fill_hist(15, lgfh, wtl);
  __fill_hist(16, lgfh, wth);

  //----- jet cross section / sigma(fcut) / -----
  user1d_dis::physfillh(17, lgfh, wtc);
  if(itype == amplitude_dis::real) user1d_dis::physfillh(17, lgfl, -wtc);

  user1d_dis::physfillh(18, lgfh, wtl);
  if(itype == amplitude_dis::real) user1d_dis::physfillh(18, lgfl, -wtl);
  
  user1d_dis::physfillh(19, lgfh, wth);
  if(itype == amplitude_dis::real) user1d_dis::physfillh(19, lgfl, -wth);
}

void UserDIS::boost_back_to_lab(const event_dis& p)
{ 
  double x = p[-1].T()/p[hadron(0)].T();
  double bz = (1.0 - x)/(1.0 + x);
  threevector<double> bVec = -((pbreit[-1] + pbreit[hadron(0)]).boostVector());
  lorentzvector<double> p0(pbreit[hadron(0)]);
  
  p0.boost(bVec);
  double phi = p0.phi(), theta = p0.theta();
  unsigned int njet = pjl.upper();
  
  for(unsigned int i = 1; i <= njet; i++) {
    pjl[i].boost(bVec);
    pjl[i].rotateZ(-phi);
    pjl[i].rotateY(-theta); 
    pjl[i].boost(0.0, 0.0, bz);
  }
}


void UserDIS::thrust(const event_dis& p, double& thrust, threevector<double>& axis)
{
  double Ecur = 0.0;
  unsigned int ncur = 0, np = p.upper();
  threevector<double> Qvec = p[-1] - p[-2];
  double Q = abs(Qvec.Z()); Qvec /= Q;
  // the momenta in the current hemisphere
  bounded_vector<lorentzvector<double> > pcur; pcur.resize(1, (int) np);
  for(unsigned int i = 1; i <= np; i++) {
	if(p[i].Z()*Qvec.Z() > 0.0) {
	  ncur++;
	  pcur[ncur] = p[i];
	  Ecur += p[i].T();
	};
  }

  // ---- return with T = 0 if current hemishpere empty ----
  if(ncur == 0) {thrust = 0.0; axis = Qvec; return;}

  //double norm = Q;
  double norm = Ecur;
  norm /= 2.0;
  
  threevector<double> s, n0, nmax;
  if(ncur == 1) {
	thrust = Ecur/norm; s = pcur[1];
  	if(s*Qvec < 0.0) {axis = -s.unit();} else {axis = s.unit();} return;
  }
   
  double T0 = 0.0, Tmax = -1.0; 
  //     all the possible permutation for the trial thrust axises
  static const double e2[2][2] = {{1, 1}, {1,-1}};
  static const double e3[4][3] =
  {{1.0, 1.0, 1.0}, {1.0, 1.0,-1.0}, {1.0,-1.0, 1.0}, {1.0,-1.0,-1.0}};
  static const double e4[8][4] =
  {{1.0, 1.0, 1.0, 1.0}, {1.0, 1.0, 1.0,-1.0}, {1.0, 1.0,-1.0, 1.0},
   {1.0,-1.0, 1.0, 1.0}, {1.0, 1.0,-1.0,-1.0}, {1.0,-1.0, 1.0,-1.0},
   {1.0,-1.0,-1.0, 1.0}, {1.0,-1.0,-1.0,-1.0}};
  
  unsigned int nperm = 1<<(ncur-1); //2^(ncur-1)
  for(unsigned it = 0; it < nperm; it++) {
	switch(ncur) {
	  case 2: s = e2[it][0]*pcur[1] + e2[it][1]*pcur[2]; break;
	  case 3: s = e3[it][0]*pcur[1] + e3[it][1]*pcur[2]
			  + e3[it][2]*pcur[3]; break;
	  case 4: s = e4[it][0]*pcur[1] + e4[it][1]*pcur[2]
			  + e4[it][2]*pcur[3] + e4[it][3]*pcur[4]; break;
	  default: throw "the thrust is defined only for n <= 4 cases"; break;
	}
	//----- the thrust is a unit vector -----
	n0 = s.unit();
	
	//----- unnormalized thrust value ---- 
	T0 = s.mag();
		
	if(T0 > Tmax) {
	  Tmax = T0;
	  nmax = n0;
	}
  }
  
  //---- return values ----
  thrust = Tmax/norm;
  if(nmax*Qvec < 0.0) {axis = -nmax;} else {axis = nmax;}
}

void UserDIS::thrustM(const event_dis& p, double& thrust, threevector<double>& axis)
{
  double Etot = 0.0;
  unsigned int np = p.upper();
  threevector<double> Qvec = p[-1] - p[-2];
  double Q = abs(Qvec.Z()); Qvec /= Q;
  // the momenta projected to the plane perpendicular to the beam
  bounded_vector<lorentzvector<double> > perp; perp.resize(1, (int) np);
  for(unsigned int i = 1; i <= np; i++) {
	perp[i] = p[i];
	perp[i].Z() = 0.0;
	Etot += p[i].T();
	};

  double norm = Q/2.0;
  //double norm = Etot;
  
  threevector<double> s, n0, nmax;
   
  double T0 = 0.0, Tmax = -1.0; 
  //     all the possible permutation for the trial thrust axises
  static const double e2[2][2] = {{1, 1}, {1,-1}};
  static const double e3[4][3] =
  {{1.0, 1.0, 1.0}, {1.0, 1.0,-1.0}, {1.0,-1.0, 1.0}, {1.0,-1.0,-1.0}};
  static const double e4[8][4] =
  {{1.0, 1.0, 1.0, 1.0}, {1.0, 1.0, 1.0,-1.0}, {1.0, 1.0,-1.0, 1.0},
   {1.0,-1.0, 1.0, 1.0}, {1.0, 1.0,-1.0,-1.0}, {1.0,-1.0, 1.0,-1.0},
   {1.0,-1.0,-1.0, 1.0}, {1.0,-1.0,-1.0,-1.0}};
  
  unsigned int nperm = 1<<(np-1); //2^(np-1)
  for(unsigned it = 0; it < nperm; it++) {
	switch(np) {
	  case 2: s = e2[it][0]*perp[1] + e2[it][1]*perp[2]; break;
	  case 3: s = e3[it][0]*perp[1] + e3[it][1]*perp[2]
			  + e3[it][2]*perp[3]; break;
	  case 4: s = e4[it][0]*perp[1] + e4[it][1]*perp[2]
			  + e4[it][2]*perp[3] + e4[it][3]*perp[4]; break;
	  default: throw "the thrust is defined only for n <= 4 cases"; break;
	}
	//----- the thrust is a unit vector -----
	n0 = s.unit();
	
	//----- unnormalized thrust value ---- 
	T0 = s.mag();
		
	if(T0 > Tmax) {
	  Tmax = T0;
	  nmax = n0;
	}
  }
  
  //---- return values ----
  thrust = Tmax/norm;
  axis = nmax;
}
