//------ DON'T TOUCH THIS PART! ------
#include <bits/photo-phasespace.h>
#include <bits/photo-process.h>
#include <bits/photo-jetfunc.h>
#include <nlo++-module_add.h>

//----- used namespaces -----
using namespace nlo;
using namespace std;


//----- declaration of the user defined functons -----
void inputfunc(unsigned int&, unsigned int&, unsigned int&);
void psinput(phasespace_photo *, double&, double&);
user_base_photo * userfunc();

typedef unsigned long int (*module_add_type)(bool, const list<basic_string<char> >&, const basic_string<char>&);
extern  module_add_type module_add;


//----- array of the symbols symbols -----
extern "C"{
  
  struct { 
	const char *name;
	void *address;
  } user_defined_functions[] = 
  {
	//   process index: hhc for gamma + p --> jets (photoproduction direct photon)
	{"procindex", (void *) "photodir"},
  
	//   input function 
	{"inputfunc", (void *) inputfunc},
  
	//   phase space input function 
	{"psinput", (void *) psinput},
  
	//   user defined functions
	{"userfunc",  (void *) userfunc},
  
	//   module to generate the readable result
	{"main_module_add", (void *) module_add},
  
	//  end of the list
	{0, 0}
  };
}
//------ END OF THE DO-NOT-TOUCH-PART ------



//------ USER DEFINED PART STARTS HERE ------
#include <algorithm>
#include <kT_clus.h>
#include "pdf-cteq6.h"


//    Defines new sample type to do felavor decomposation in the initial state
const char *mySample_label[4] = {"g", "u", "d", "total"};
typedef weight<4U, mySample_label> mySample;

namespace nlo {
  template<>
  struct weight_conversion<weight_photo, mySample>
    : public std::unary_function<weight_photo, mySample>
  {
    mySample operator()(const weight_photo& x)
    {
      mySample res;
      
      for(unsigned int i = 0; i < 3; i++)
		res[3] += (res[i] = x[i]);

      return res;
    }
  };
}

//   Here we specify some base user objects with the new sample type
//     (see photo-jetfunc.h and included header files therein) 
typedef basic_user<jetfunc_photo, void,        mySample, weight_conversion> myUser0d;
typedef basic_user<jetfunc_photo, double,      mySample, weight_conversion> myUser1d;
typedef basic_user<jetfunc_photo, histpoint1d, mySample, weight_conversion> myUser1h;
typedef basic_user<jetfunc_photo, histpoint2d, mySample, weight_conversion> myUser2h;


class UserPhoto
#if defined(_SCALES_)
  : public myUser1d
#elif defined(_XSECTION_)
  : public basic_user_set<myUser0d, myUser1h, myUser2h> 
#endif
{
public:
  //   init and user function
  void initfunc(unsigned int);
  void userfunc(const event_photo&, const amplitude_photo&);
  
private:
  //   pdf
  pdf_cteq6 pdf;

  // algorithms
  kT_clus_long jetclus;
   
  //  private types
  typedef lorentzvector<double> _Lv;

  // the jet structore
  bounded_vector<_Lv> cj, pj_lab, pj_3jet; 
  bounded_vector<unsigned int> jet;
  
  struct pT_sort {
    bool operator()(const _Lv& p1, const _Lv& p2) const {
      return p1.perp2() > p2.perp2();
    }
  };

  struct E_sort {
    bool operator()(const _Lv& p1, const _Lv& p2) const {
      return p1.T() > p2.T();
    }
  };

  //   making simpler the life
  double hard_scale_renorm, hard_scale_fact;
  const amplitude_photo *amplitude_of_event;
  

#if defined(_SCALES_) 
  //   calculate the weight in the term of scales
  weight_photo _M_renorm(const double& x) const {
    return amplitude_of_event -> operator()(x*x*hard_scale_renorm, hard_scale_fact);
  }
  
  weight_photo _M_fact(const double& x) const {
    return amplitude_of_event -> operator()(hard_scale_renorm, x*x*hard_scale_fact);
  }
  
  weight_photo _M_renorm_and_fact(const double& x) const {
    return amplitude_of_event -> operator()(x*x*hard_scale_renorm, x*x*hard_scale_fact);
  }
  
  //   helper classes
  pointer_to_unary_const_member_function<UserPhoto, const double&, weight_photo>
  renorm, fact, renorm_fact;
  
public:
  //   default constructor
  UserPhoto()
    : renorm(this, &UserPhoto::_M_renorm), 
      fact(this, &UserPhoto::_M_fact),
      renorm_fact(this, &UserPhoto::_M_renorm_and_fact) {}  
  
#elif defined(_XSECTION_)
  void fill_jets(int, double);
#endif
};
  

//----- defines the module to sum up the results of the different runs -----
module_add_type module_add = 
#if defined(_SCALES_)
main_module_add<basic_user_result<myUser1d::distbook_type> >;
#elif defined(_XSECTION_)
main_module_add<basic_user_result<myUser0d::distbook_type, myUser1h::distbook_type, myUser2h::distbook_type> >; 
#endif


user_base_photo * userfunc() {
  return new UserPhoto;
}

void inputfunc(unsigned int& nj, unsigned int& nu, unsigned int& nd)
{
  //  number of jets
  nj = 3U;
  
  //  number of the up and down type flavours
  nu = 2U;
  nd = 3U;
} 

void psinput(phasespace_photo *ps, double& el, double& eh)
{
  //  energy of incomings
  el = 27.5;
  eh = 820.0;
  
  //   You can use your own phase generator. 
  //   Here we use the default.
  ps = 0;
} 


void UserPhoto::initfunc(unsigned int)
{
#if defined(_SCALES_)
  phys(1, "kT renorm scale dep.",        60, 0.1, 2.0);
  phys(2, "kT fact scale dep.",          60, 0.1, 2.0);
  phys(3, "kT renorm & fact scale dep.", 60, 0.1, 2.0);
#elif defined(_XSECTION_)
  myUser0d::phys(0, "sigma(pT(1,2) > 6GeV, pT(3) > 5GeV) xmu = 1.0");

  myUser1h::phys(1, "M3J distribucion xmu = 0.5", 20, 50.0, 150.0, 0.2);
  myUser1h::phys(2, "M3J distribucion xmu = 1.0", 20, 50.0, 150.0, 0.2);
  myUser1h::phys(3, "M3J distribucion xmu = 2.0", 20, 50.0, 150.0, 0.2);

  myUser1h::phys(11, "pT(1) distribucion xmu = 0.5", 35, 6.0, 76.0, 0.2);
  myUser1h::phys(12, "pT(1) distribucion xmu = 1.0", 35, 6.0, 76.0, 0.2);
  myUser1h::phys(13, "pT(1) distribucion xmu = 2.0", 35, 6.0, 76.0, 0.2);
  
  myUser1h::phys(21, "X3 Dalitz variable xmu = 1.0", 20,  0.5, 1.0, 0.2);
  myUser1h::phys(22, "X4 Dalitz variable xmu = 1.0", 20,  0.5, 1.0, 0.2);
  myUser1h::phys(23, "cos(theta3) angle  xmu = 1.0", 40, -1.0, 1.0, 0.2);
  myUser1h::phys(24, "cos(psi3) angle    xmu = 1.0", 40, -1.0, 1.0, 0.2);

  myUser2h::phys(0, "dsigma/(dX3dX4) distribution", 20, 0.5, 1.0,  20, 0.5, 1.0);
#endif
}


void UserPhoto::userfunc(const event_photo& p, const amplitude_photo& amp)
{
  //----- initialize the user object -----
  amplitude_of_event = &amp;

  //----- photon momentum fraction -----
  double y = (p[-1]*p[hadron(0)])/(p[hadron(-1)]*p[hadron(0)]);
  if(y < 0.2 || y > 0.8) return; 

  //----- do the cluster analysis-----
  jetclus.set_up(1,false,2);
  jetclus(p); jetclus.incl(cj, jet);
  
  unsigned int nj = cj.upper();
  if(nj < 3) return;   //  unable to resolve three jets

  //----- ZEUS cuts -----
  double pT, hT = 0.0;
  double pTmin = 5.0, emax = 2.4;
  
  pj_lab.resize(1,0);
  for(unsigned int i = 1; i <= nj; i++)
    if((pT = cj[i].perp()) > pTmin && abs(cj[i].prapidity()) < emax) {
      hT += pT;
      pj_lab.push_back(cj[i]);
    }
  
  nj = pj_lab.upper();
  if(nj < 3 || hT < 17.0) return;   //  unable to resolve three jets
  
  //----- sorting the jets by transverse energy -----
  std::sort(pj_lab.begin(), pj_lab.end(), pT_sort()); 
  if(pj_lab[1].perp() < 6.0 || pj_lab[2].perp() < 6.0) return;
  
  //----- 3-jet mass -----
  lorentzvector<double> q = pj_lab[1]+pj_lab[2]+pj_lab[3];
  double MJ = q.mag();
  
  //----- ZEUS cuts -----
  if(MJ < 50.0) return;
  
  //----- computes the angular variables -----
  threevector<double> pav = p[-1]-p[0];
  if(p[-1].T() < p[0].T()) pav = -pav;
  
  double cosT3 = cosAngle(pav, pj_lab[1]);
  if(abs(cosT3) > 0.8) return;

  //----- 3-jet rest frame -----
  threevector<double> bVec = -q.boostVector();
  
  pj_3jet = pj_lab; pj_3jet[1].boost(bVec);
  pj_3jet[2].boost(bVec); pj_3jet[3].boost(bVec);
  
  //----- sorting the three leading jets by energy -----
  std::sort(pj_3jet.begin(), pj_3jet.begin()+3, E_sort());
  
  //----- Dalitz variables and 3-jet angles -----
  double X3 = 2.0*pj_3jet[1].T()/MJ;
  
  //----- ZEUS cuts -----
  if(X3 > 0.95) return;

  //----- pdf -----
  amp.pdf_and_qcd_coupling(pdf, 389385730.0/137.0);
  
  //----- hard scale -----
  hard_scale_renorm = hard_scale_fact = pj_lab[1].perp2();
  
#if defined(_SCALES_)
  //----- calculate the scale dependences -----
  physfill(1, renorm);
  physfill(2, fact);
  physfill(3, renorm_fact);
  
#elif defined(_XSECTION_)
  double X4 = 2.0*pj_3jet[2].T()/MJ;
  double cosP3 = cosAngle(cross(pj_lab[1],pav), cross(pj_lab[2],pj_lab[3]));
  
  //----- calculate the weight -----
  amplitude_photo::contrib_type itype = amp.contrib();
  const weight_photo& wt = amp(hard_scale_renorm, hard_scale_fact);

  //----- total 3-jet cross section -----
  myUser0d::physfill(0, wt);

  //----- fill the pT and MJ distributions -----
  fill_jets(1, 0.5);
  fill_jets(2, 1.0);
  fill_jets(3, 2.0);
 
  //----- Dalitz plots -----  
  switch(itype) {
  case amplitude_photo::real: sphysfilld(21, X3, sphysreal(wt)); break;
  case amplitude_photo::sub:  sphysfilld(21, X3, sphyssub(wt));  break;
  default: physfilld(21, X3, wt);
  }
  
  switch(itype) {
  case amplitude_photo::real: sphysfilld(22, X4, sphysreal(wt)); break;
  case amplitude_photo::sub:  sphysfilld(22, X4, sphyssub(wt));  break;
  default: physfilld(22, X4, wt);
  }
  
  //----- two dimensional Dalitz plot -----
  physfilldd(0, X3, X4, wt);

  //----- 3-jet anlgles -----
  switch(itype) {
  case amplitude_photo::real: sphysfilld(23, cosT3, sphysreal(wt)); break;
  case amplitude_photo::sub:  sphysfilld(23, cosT3, sphyssub(wt));  break;
  default: physfilld(23, cosT3, wt);
  }
  
  switch(itype) {
  case amplitude_photo::real: sphysfilld(24, cosP3, sphysreal(wt)); break;
  case amplitude_photo::sub:  sphysfilld(24, cosP3, sphyssub(wt));  break;
  default: physfilld(24, cosP3, wt);
  }
#endif
}



#if defined(_XSECTION_)
void UserPhoto::fill_jets(int id, double x)
{ 
  //----- calculate the weight -----
  amplitude_photo::contrib_type itype = amplitude_of_event -> contrib();
  const weight_photo& wt = amplitude_of_event -> operator()(x*x*hard_scale_renorm, x*x*hard_scale_fact);
  
  //----- 3-jet mass distribution -----
  double MJ = (pj_lab[1]+pj_lab[2]+pj_lab[3]).mag();

  switch(itype) {
  case amplitude_photo::real: sphysfilld(id, MJ, sphysreal(wt)); break;
  case amplitude_photo::sub:  sphysfilld(id, MJ, sphyssub(wt));  break;
  default: physfilld(id, MJ, wt);
  }

  //----- leading jet pT distribution -----
  double pT = pj_lab[1].perp();
  
  switch(itype) {
  case amplitude_photo::real: sphysfilld(10+id, pT, sphysreal(wt)); break;
  case amplitude_photo::sub:  sphysfilld(10+id, pT, sphyssub(wt));  break;
  default: physfilld(10+id, pT, wt);
  }
}
#endif
