//------ DON'T TOUCH THIS PART! ------
#include <bits/hhc2ph-phasespace.h>
#include <bits/hhc2ph-process.h>
#include <bits/hhc2ph-jetfunc.h>
#include <nlo++-module_add.h>

//----- used namespaces -----
using namespace nlo;
using namespace std;


//----- declaration of the user defined functons -----
void inputfunc(unsigned int&, unsigned int&, unsigned int&);
void psinput(phasespace_hhc2ph *, double&, double&, double&);
user_base_hhc2ph * userfunc();

typedef unsigned long int (*module_add_type)(bool, const list<basic_string<char> >&, const basic_string<char>&);
extern  module_add_type module_add;


//----- array of the symbols symbols -----
extern "C"{
  
  struct { 
	const char *name;
	void *address;
  } user_defined_functions[] = 
  {
	//   process index: hhc for hadron-hadron --> 2 photons + jets
	{"procindex", (void *) "hhc2ph"},
  
	//   input function 
	{"inputfunc", (void *) inputfunc},
  
	//   phase space input function 
	{"psinput", (void *) psinput},
  
	//   user defined functions
	{"userfunc",  (void *) userfunc},
  
	//   module to generate the readable result
	{"main_module_add", (void *) module_add},
  
	//  end of the list
	{0, 0}
  };
}
//------ END OF THE DO-NOT-TOUCH-PART ------


//------ USER DEFINED PART STARTS HERE ------
#include <algorithm>
#include <bits/phys-cone_seedless.h>
#include "pdf-cteq6.h"



class UserHHC2PH : public user1h_hhc2ph
{
public:
  //   init and user function
  void initfunc(unsigned int);
  void userfunc(const event_hhc2ph&, const amplitude_hhc2ph&);
  
private:
  //   pdf
  pdf_cteq6 pdf;

  // algorithms
  cone_seedless jetclus;
  
  // the jet structure
  bounded_vector<lorentzvector<double> > cj, pj; 
  bounded_vector<unsigned int> jet;

  //   fill the histograms
  void fill_hist(int, amplitude_hhc2ph::contrib_type, double, const weight_hhc2ph&);
  
  //   izolation cone
  struct izolation_cone {
    double pT, R;
  };

  //   lists of the izolation cones
  bounded_vector<izolation_cone> cone1, cone2;

  //   order the cones
  struct radius_sort {
    bool operator()(const izolation_cone& c1, const izolation_cone& c2) const {
      return c1.R < c2.R;
    }
  };

  //   do the photon izolation
  bool photon_izolation(const event_hhc2ph&, double, double=1.0, unsigned int=1U);
  
  //   calculate the radius between two particles
  double radius(const lorentzvector<double>&, const lorentzvector<double>&);
};


//----- defines the module to sum up the results of the different runs -----
module_add_type module_add =  main_module_add<basic_user_result<user1h_hhc2ph::distbook_type> >;


user_base_hhc2ph * userfunc() {
  return new UserHHC2PH;
}


void inputfunc(unsigned int& nj, unsigned int& nu, unsigned int& nd)
{
  //  number of jets
  nj = 1U;
  
  //  number of the up and down type flavours
  nu = 2U;
  nd = 3U;
} 

void psinput(phasespace_hhc2ph *ps, double& s, double& q2min, double& q2max)
{
  //   we use the default phase space generator
  ps = 0;
  
  // total c.m energy square
  s = 14000*14000;
  
  //  Q^2 cuts
  q2min = 80.0*80.0;  // GeV^2
  q2max = 160.0*160.0;  // GeV^2 
}


void UserHHC2PH::initfunc(unsigned int)
{  
  phys(0, "Mgg dist Rg = 0.4 eps = 1.0",30, 75.0, 165.0, 0.2);
  phys(1, "Mgg dist Rg = 0.7 eps = 1.0",30, 75.0, 165.0, 0.2);
  phys(2, "Mgg dist Rg = 1.0 eps = 1.0",30, 75.0, 165.0, 0.2);
  phys(3, "Mgg dist Rg = 0.4 eps = 0.5",30, 75.0, 165.0, 0.2);
  phys(4, "Mgg dist Rg = 0.7 eps = 0.5",30, 75.0, 165.0, 0.2);
  phys(5, "Mgg dist Rg = 1.0 eps = 0.5",30, 75.0, 165.0, 0.2);
  phys(6, "Mgg dist Rg = 0.4 eps = 0.1",30, 75.0, 165.0, 0.2);
  phys(7, "Mgg dist Rg = 0.7 eps = 0.1",30, 75.0, 165.0, 0.2);
  phys(8, "Mgg dist Rg = 1.0 eps = 0.1",30, 75.0, 165.0, 0.2);

}


void UserHHC2PH::userfunc(const event_hhc2ph& p, const amplitude_hhc2ph& amp)
{
  typedef lorentzvector<double> _Lv;
  
  //    no photon belongs the beam
  double fTmin = 40, femax = 2.5;

  //----- photon pT -----
  double pT1 = p[-3].perp(), pT2 = p[-2].perp(), pT12 = (p[-3]+p[-2]).perp();

  if(pT1 < fTmin || pT2 < fTmin || pT12 < 40.0) return;
  if(abs(p[-3].prapidity()) > femax || abs(p[-2].prapidity()) > femax) return;
 
  //  photon izolation
  bool izo0 = photon_izolation(p, 0.4, 1.0, 1U);
  bool izo1 = photon_izolation(p, 0.7, 1.0, 1U);
  bool izo2 = photon_izolation(p, 1.0, 1.0, 1U);
  bool izo3 = photon_izolation(p, 0.4, 0.5, 1U);
  bool izo4 = photon_izolation(p, 0.7, 0.5, 1U);
  bool izo5 = photon_izolation(p, 1.0, 0.5, 1U);
  bool izo6 = photon_izolation(p, 0.4, 0.1, 1U);
  bool izo7 = photon_izolation(p, 0.7, 0.1, 1U);
  bool izo8 = photon_izolation(p, 1.0, 0.1, 1U);

  if(izo0 == false && izo1 == false && izo2 == false && 
     izo3 == false && izo4 == false && izo5 == false &&
     izo6 == false && izo7 == false && izo8 == false) return;
  
  //----- do the cluster analysis-----
  event_hhc2ph::const_iterator p1 = p.begin()+2;
  event_hhc2ph::const_iterator pn = p.end();
  
  if(p[p.upper()].perp() < 1e-9) pn--;
  
  cj.resize(1,0); 
  if((unsigned int) (pn-p1) == 1U) cj.push_back(p[1]);
  else cj = jetclus(p1, pn, 1.0);
  
  unsigned int nj = cj.upper();
  if(nj < 1) return;   //  unable to resolve at least one jet
  
  //----- cuts -----
  double pT, hT = 0.0;
  double pTmin = 40.0, emax = 2.5;
  
  pj.resize(1,0);
  for(unsigned int i = 1; i <= nj; i++)
    if((pT = cj[i].perp()) > pTmin && abs(cj[i].prapidity()) < emax) {
      hT += pT;
      pj.push_back(cj[i]);
    }
  
  nj = pj.upper();
  if(nj < 1) return;   //  unable to resolve at least one jet
  
  //----- pdf -----
  if(amp.integral() == amplitude_hhc2ph::lo) pdf.mode(4);
  else pdf.mode(1);
  amp.pdf_and_qcd_coupling(pdf, 389385730000.0/(137.0*137.0));
  
  //----- type of the contribution -----
  weight_hhc2ph wt;
  double mu2, Mff = sqrt(2.0*(p[-3]*p[-2]));
  amplitude_hhc2ph::contrib_type itype = amp.contrib();
    
  //----- calculate & fill the weight -----
  double r1 = radius(pj[1],p[-3]);
  double r2 = radius(pj[1],p[-2]);
  if(pT1 < pT2) std::swap(r1, r2);

  if(r1 > 1.5 && r2 > 1.5) {
    pT = pj[1].perp();
    mu2 = (Mff*Mff + pT*pT)/4.0; 
    wt = amp(mu2,mu2);
    
    if(izo0 == true) fill_hist(0, itype, Mff, wt);
    if(izo1 == true) fill_hist(1, itype, Mff, wt);
    if(izo2 == true) fill_hist(2, itype, Mff, wt);
    if(izo3 == true) fill_hist(3, itype, Mff, wt);
    if(izo4 == true) fill_hist(4, itype, Mff, wt);
    if(izo5 == true) fill_hist(5, itype, Mff, wt);
    if(izo6 == true) fill_hist(6, itype, Mff, wt);
    if(izo7 == true) fill_hist(7, itype, Mff, wt);
    if(izo8 == true) fill_hist(8, itype, Mff, wt);
  }
  
  if(nj == 1) return;
  
  //----- contribution of the possible second jet -----
  r1 = radius(pj[2],p[-3]);
  r2 = radius(pj[2],p[-2]);
  if(pT1 < pT2) std::swap(r1, r2);
  
  if(r1 > 1.5 && r2 > 1.5) {
    pT = pj[2].perp();
    mu2 = (Mff*Mff + pT*pT)/4.0; 
    wt = amp(mu2,mu2);
    
    if(izo0 == true) fill_hist(0, itype, Mff, wt);
    if(izo1 == true) fill_hist(1, itype, Mff, wt);
    if(izo2 == true) fill_hist(2, itype, Mff, wt);
    if(izo3 == true) fill_hist(3, itype, Mff, wt);
    if(izo4 == true) fill_hist(4, itype, Mff, wt);
    if(izo5 == true) fill_hist(5, itype, Mff, wt);
    if(izo6 == true) fill_hist(6, itype, Mff, wt);
    if(izo7 == true) fill_hist(7, itype, Mff, wt);
    if(izo8 == true) fill_hist(8, itype, Mff, wt);
  }
}

void UserHHC2PH::
fill_hist(int id, amplitude_hhc2ph::contrib_type itype, double pT, const weight_hhc2ph& wt) 
{
  switch(itype) {
  case amplitude_hhc2ph::real: sphysfilld(id, pT, sphysreal(wt)); break;
  case amplitude_hhc2ph::sub:  sphysfilld(id, pT, sphyssub(wt));  break;
  default: physfilld(id, pT, wt);
  }
}

double UserHHC2PH::
radius(const lorentzvector<double>& p1, const lorentzvector<double>& p2)
{
  static const double pi = 3.14159265358979323846;
  static const double twopi = 6.28318530717958647692;

  double deta = p1.rapidity() - p2.rapidity();
  double dphi = p1.phi() - p2.phi();

  if(dphi >= pi) dphi = fmod(pi+dphi, twopi) - pi;
  else if(dphi < -pi) dphi = -fmod(pi-dphi, twopi) + pi;

  return sqrt(deta*deta+dphi*dphi);
}


#define Xi(E, x) (eps*(E)*std::pow((1.0-std::cos(x))/(1.0-std::cos(delta)), (int) n))

bool UserHHC2PH::
photon_izolation(const event_hhc2ph& p, double delta, double eps, unsigned int n)
{
  unsigned int np = p.upper();
  
  if(p[np].perp() < 1e-9) np--;
  cone1.resize(1,np); cone2.resize(1,np); 
  
  for(unsigned int i = 1; i <= np; i++) {
    cone1[i].pT = cone2[i].pT = p[i].perp();
    cone1[i].R = radius(p[-3], p[i]);
    cone2[i].R = radius(p[-2], p[i]);
  }
  
  if(np > 1) {
    std::sort(cone1.begin(), cone1.end(), radius_sort());
    std::sort(cone2.begin(), cone2.end(), radius_sort());
  }

  double Esum1 = 0.0, Esum2 = 0.0;
  double pT1 = p[-3].perp(), pT2 = p[-2].perp();
  for(unsigned int i = 1; i <= np; i++) {
    Esum1 += delta > cone1[i].R ? cone1[i].pT : 0.0;
    Esum2 += delta > cone2[i].R ? cone2[i].pT : 0.0;
    if(Xi(pT1, (delta < cone1[i].R ? delta : cone1[i].R)) < Esum1) return false;
    if(Xi(pT2, (delta < cone2[i].R ? delta : cone2[i].R)) < Esum2) return false;
  }

  return true;
}
