//------ DON'T TOUCH THIS PART! ------
#include <bits/hhc2ph-phasespace.h>
#include <bits/hhc2ph-process.h>
#include <bits/hhc2ph-jetfunc.h>
#include <nlo++-module_add.h>

//----- used namespaces -----
using namespace nlo;
using namespace std;


//----- declaration of the user defined functons -----
void inputfunc(unsigned int&, unsigned int&, unsigned int&);
void psinput(phasespace_hhc2ph *, double&, double&, double&);
user_base_hhc2ph * userfunc();

typedef unsigned long int (*module_add_type)(bool, const list<basic_string<char> >&, const basic_string<char>&);
extern  module_add_type module_add;


//----- array of the symbols symbols -----
extern "C"{
  
  struct { 
	const char *name;
	void *address;
  } user_defined_functions[] = 
  {
	//   process index: hhc for hadron-hadron --> 2 photons + jets
	{"procindex", (void *) "hhc2ph"},
  
	//   input function 
	{"inputfunc", (void *) inputfunc},
  
	//   phase space input function 
	{"psinput", (void *) psinput},
  
	//   user defined functions
	{"userfunc",  (void *) userfunc},
  
	//   module to generate the readable result
	{"main_module_add", (void *) module_add},
  
	//  end of the list
	{0, 0}
  };
}
//------ END OF THE DO-NOT-TOUCH-PART ------


//------ USER DEFINED PART STARTS HERE ------
#include <algorithm>
#include <bits/phys-cone_seedless.h>
#include "pdf-cteq6.h"


class UserHHC2PH : public user1d_hhc2ph
{
public:
  //   init and user function
  void initfunc(unsigned int);
  void userfunc(const event_hhc2ph&, const amplitude_hhc2ph&);
  
private:
  //   pdf
  pdf_cteq6 pdf;

  // algorithms
  cone_seedless jetclus;
  
  // the jet structure
  bounded_vector<lorentzvector<double> > cj, pj; 
  bounded_vector<unsigned int> jet;
 
  //   izolation cone
  struct izolation_cone {
    double pT, R;
  };

  //   lists of the izolation cones
  bounded_vector<izolation_cone> cone1, cone2;

  //   order the cones
  struct radius_sort {
    bool operator()(const izolation_cone& c1, const izolation_cone& c2) const {
      return c1.R < c2.R;
    }
  };

  //  scale studies
  class physscales
  {
  public:
    //   constructor
    physscales() {}
    physscales(const amplitude_hhc2ph& aa, double hs)
      : hard_scale(hs), amp(&aa) {}

    //   set the input parameters
    void operator()(const amplitude_hhc2ph& aa, double hs) {
      amp = &aa; hard_scale = hs;
    }

    //   calculate the weight in the term of scales
    weight_hhc2ph renorm(const double& x) const {
      return amp -> operator()(x*x*hard_scale, hard_scale);
    }

    weight_hhc2ph fact(const double& x) const {
      return amp -> operator()(hard_scale, x*x*hard_scale);
    }

    weight_hhc2ph renorm_and_fact(const double& x) const {
      return amp -> operator()(x*x*hard_scale, x*x*hard_scale);
    }
    
  private:
    //   private members
    double hard_scale;
    const amplitude_hhc2ph *amp;
  };
  
  //   helper classes
  physscales scalefunc;
  
  pointer_to_unary_const_member_function<physscales, const double&, weight_hhc2ph>
  renorm, fact, renorm_fact;
  
public:
  //   default constructor
  UserHHC2PH()
    : renorm(scalefunc, &physscales::renorm), fact(scalefunc, &physscales::fact),
      renorm_fact(scalefunc, &physscales::renorm_and_fact) {}  

private:  
  //   do the photon izolation
  bool photon_izolation(const event_hhc2ph&, double, double=1.0, unsigned int=1U);
  
  //   calculate the radius between two particles
  static double radius(const lorentzvector<double>&, const lorentzvector<double>&);
};


//----- defines the module to sum up the results of the different runs -----
module_add_type module_add =  main_module_add<basic_user_result<user1h_hhc2ph::distbook_type> >;


user_base_hhc2ph * userfunc() {
  return new UserHHC2PH;
}

void inputfunc(unsigned int& nj, unsigned int& nu, unsigned int& nd)
{
  //  number of jets
  nj = 1U;
  
  //  number of the up and down type flavours
  nu = 2U;
  nd = 3U;
} 

void psinput(phasespace_hhc2ph *ps, double& s, double& q2min, double& q2max)
{
  //   we use the default phase space generator
  ps = 0;
  
  // total c.m energy square
  s = 14000*14000;
  
  //  Q^2 cuts
  q2min = 118.5*118.5;  // GeV^2
  q2max = 121.5*121.5;  // GeV^2 
}


void UserHHC2PH::initfunc(unsigned int)
{  
  phys(1, "cone renorm scale dep. Rg = 0.4 epsilon = 0.5",        60, 0.1, 2.0);
  phys(2, "cone fact scale dep. Rg = 0.4 epsilon = 0.5",          60, 0.1, 2.0);
  phys(3, "cone renorm & fact scale dep. Rg = 0.4 epsilon = 0.5", 60, 0.1, 2.0);

  phys(11, "cone renorm scale dep. Rg = 1.0 epsilon = 0.1",        60, 0.1, 2.0);
  phys(12, "cone fact scale dep. Rg = 1.0 epsilon = 0.1",          60, 0.1, 2.0);
  phys(13, "cone renorm & fact scale dep. Rg = 1.0 epsilon = 0.1", 60, 0.1, 2.0);
}


void UserHHC2PH::userfunc(const event_hhc2ph& p, const amplitude_hhc2ph& amp)
{
  typedef lorentzvector<double> _Lv;
  
  //    no photon belongs the beam
  double fTmin = 40, femax = 2.5;
  double pT1 = p[-3].perp(), pT2 = p[-2].perp(), pT12 = (p[-3]+p[-2]).perp();

  if(pT1 < fTmin || pT2 < fTmin || pT12 < 40.0) return;
  if(abs(p[-3].prapidity()) > femax || abs(p[-2].prapidity()) > femax) return;

  //  photon izolation
  bool iso1 = photon_izolation(p, 0.4, 0.5, 1U);
  bool iso2 = photon_izolation(p, 1.0, 0.1, 1U);
  if(iso1 == false && iso2 == false) return;
  
  //----- do the cluster analysis-----
  event_hhc2ph::const_iterator p1 = p.begin()+2;
  event_hhc2ph::const_iterator pn = p.end();

  if(p[p.upper()].perp() < 1e-9) pn--;
  
  cj.resize(1,0); 
  if((unsigned int) (pn-p1) == 1U) cj.push_back(p[1]);
  else cj = jetclus(p1, pn, 1.0);
  
  unsigned int nj = cj.upper();
  if(nj < 1) return;   //  unable to resolve at least one jet
  
  //----- cuts -----
  double pT, hT = 0.0;
  double pTmin = 40.0, emax = 2.5;
  
  pj.resize(1,0);
  for(unsigned int i = 1; i <= nj; i++)
    if((pT = cj[i].perp()) > pTmin && abs(cj[i].prapidity()) < emax) {
      hT += pT;
      pj.push_back(cj[i]);
    }
  
  nj = pj.upper();
  if(nj < 1) return;   //  unable to resolve at least one jet
  
  //----- pdf -----
  if(amp.integral() == amplitude_hhc2ph::lo) pdf.mode(4);
  else pdf.mode(1);
  amp.pdf_and_qcd_coupling(pdf, 389385730000.0/(137.0*137.0));
  
  //----- type of the contribution -----
  double mu2, Mff = sqrt(2.0*(p[-3]*p[-2]));

  //----- calculate & fill the weight -----
  double r1 = radius(pj[1],p[-3]);
  double r2 = radius(pj[1],p[-2]);
  
  if(r1 > 1.5 && r2 > 1.5) {
    pT = pj[1].perp(); 
    mu2 = (Mff*Mff + pT*pT)/4.0; 
    scalefunc(amp, mu2);
    
    //----- calculate the scale dependences -----
    if(iso1 == true) {
	  physfill(1, renorm);
	  physfill(2, fact);
      physfill(3, renorm_fact);
    }

    if(iso2 == true) {
	  physfill(11, renorm);
	  physfill(12, fact);
      physfill(13, renorm_fact);
    }
  }

  if(nj == 1) return;
  
  //----- contribution of the possible second jet -----
  r1 = radius(pj[2],p[-3]);
  r2 = radius(pj[2],p[-2]);
  
  if(r1 > 1.5 && r2 > 1.5) {
    pT = pj[2].perp(); 
    mu2 = (Mff*Mff  + pT*pT)/4.0; 
    scalefunc(amp, mu2);
    
    if(iso1 == true) {
	  physfill(1, renorm);
	  physfill(2, fact);
      physfill(3, renorm_fact);
    }

    if(iso2 == true) {
	  physfill(11, renorm);
	  physfill(12, fact);
      physfill(13, renorm_fact);
    }
  }
}

double UserHHC2PH::
radius(const lorentzvector<double>& p1, const lorentzvector<double>& p2)
{
  static const double pi = 3.14159265358979323846;
  static const double twopi = 6.28318530717958647692;

  double deta = p1.rapidity() - p2.rapidity();
  double dphi = p1.phi() - p2.phi();

  if(dphi >= pi) dphi = fmod(pi+dphi, twopi) - pi;
  else if(dphi < -pi) dphi = -fmod(pi-dphi, twopi) + pi;

  return sqrt(deta*deta+dphi*dphi);
}


#define Xi(E, x) (eps*(E)*std::pow((1.0-std::cos(x))/(1.0-std::cos(delta)), (int) n))

bool UserHHC2PH::
photon_izolation(const event_hhc2ph& p, double delta, double eps, unsigned int n)
{
  unsigned int np = p.upper();
  
  if(p[np].perp() < 1e-9) np--;
  cone1.resize(1,np); cone2.resize(1,np); 
  
  for(unsigned int i = 1; i <= np; i++) {
    cone1[i].pT = cone2[i].pT = p[i].perp();
    cone1[i].R = radius(p[-3], p[i]);
    cone2[i].R = radius(p[-2], p[i]);
  }
  
  if(np > 1) {
    std::sort(cone1.begin(), cone1.end(), radius_sort());
    std::sort(cone2.begin(), cone2.end(), radius_sort());
  }

  double Esum1 = 0.0, Esum2 = 0.0;
  double pT1 = p[-3].perp(), pT2 = p[-2].perp();
  for(unsigned int i = 1; i <= np; i++) {
    Esum1 += delta > cone1[i].R ? cone1[i].pT : 0.0;
    Esum2 += delta > cone2[i].R ? cone2[i].pT : 0.0;
    if(Xi(pT1, (delta < cone1[i].R ? delta : cone1[i].R)) < Esum1) return false;
    if(Xi(pT2, (delta < cone2[i].R ? delta : cone2[i].R)) < Esum2) return false;
  }

  return true;
}
