//------ DON'T TOUCH THIS PART! ------
#include <bits/hhc-phasespace.h>
#include <bits/hhc-process.h>
#include <bits/hhc-jetfunc.h>
#include <nlo++-module_add.h>

//----- used namespaces -----
using namespace nlo;
using namespace std;


//----- declaration of the user defined functons -----
void inputfunc(unsigned int&, unsigned int&, unsigned int&);
void psinput(phasespace_hhc *, double&);
user_base_hhc * userfunc();

typedef unsigned long int (*module_add_type)(bool, const list<basic_string<char> >&, const basic_string<char>&);
extern  module_add_type module_add;


//----- array of the symbols symbols -----
extern "C"{
  
  struct { 
	const char *name;
	void *address;
  } user_defined_functions[] = 
	{
	  //   process index: hhc for hadron-hadron --> jets
	  {"procindex", (void *) "hhc"},
    
	  //   input function 
	  {"inputfunc", (void *) inputfunc},

	  //   phase space input function 
	  {"psinput", (void *) psinput},
    
	  //   user defined functions
	  {"userfunc",  (void *) userfunc},
    
	  //   module to generate the readable result
	  {"main_module_add", (void *) module_add},

	  //  end of the list
	  {0, 0}
	};
}
//------ END OF THE DO-NOT-TOUCH-PART ------


//------ USER DEFINED PART STARTS HERE ------
#include "kT_clus.h"
#include "pdf-cteq6.h"


class UserHHC : public user1h_hhc
{
public:
  //   init and user function
  void initfunc(unsigned int);
  void userfunc(const event_hhc&, const amplitude_hhc&);
  
private:
  //   pdf
  pdf_cteq6 pdf;

  // algorithms
  kT_clus_long jetclus;
  
  // the macro-jet structure
  bounded_vector<lorentzvector<double> > pj; 
};

//----- defines the module to sum up the results of the different runs -----
module_add_type module_add =  main_module_add<basic_user_result<user1h_hhc::distbook_type> >;


user_base_hhc * userfunc() {
  return new UserHHC;
}

void inputfunc(unsigned int& nj, unsigned int& nu, unsigned int& nd)
{
  //  nuber of jets
  nj = 3U;
  
  //  number of the up and down type flavours
  nu = 2U;
  nd = 3U;
} 

void psinput(phasespace_hhc *ps, double& s)
{
  //  total c.m. energy square
  s = 3240000;

  //   You can use your own phase generator. 
  //   Here we use the default.
  ps = 0;
} 



void UserHHC::initfunc(unsigned int)
{
  phys(1, "kT E3*dsigma/dE3 mu = hT",   50, 0.0, 500.0, 0.25);
  phys(2, "kT E3*dsigma/dE3 mu = hT/2", 50, 0.0, 500.0, 0.25);
  phys(3, "kT E3*dsigma/dE3 mu = 2*hT", 50, 0.0, 500.0, 0.25);
}


void UserHHC::userfunc(const event_hhc& p, const amplitude_hhc& amp)
{
  typedef lorentzvector<double> _Lv;
  unsigned int nj = 3U;
  
  //----- do the cluster analysis-----    
  const bounded_vector<double>& f = jetclus(1.0, p, 1.0);
  
  //----- jet shape variable fn  -----
  double fh = sqrt(f[nj] > f[nj+1] ? f[nj] : f[nj+1]);
  
  //----- resolve the macro-jet structure -----
  pj.resize(1, (int) nj); jetclus.reco(nj, pj);
  
  //----- rapidity cut -----
  double emax = 2.0;
  double hT = 0.0, pT_max = 0.0, pT;
  unsigned int jetnum = 0;
  
  for(unsigned int i = 1; i <= nj; i++)
    if(abs(pj[i].prapidity()) < emax) {
      if((pT= pj[i].perp()) > pT_max) pT_max = pT;
      hT += pT;
      jetnum++;
    }
  
  if(jetnum != nj) return;  //  unable to resolve njet
  
  //----- type of the contribution & pdf -----
  amplitude_hhc::contrib_type itype = amp.contrib();
  amp.pdf_and_qcd_coupling(pdf, 389385.730);
  
  //----- scales -----
  double mu2 = hT*hT/9.0;
  
  //----- calculate the weight -----
  weight_hhc wt = fh*amp(mu2, mu2);
  
  //----- diff. jet cross section E3*dsigma/dE3 -----
  switch(itype) {
  case amplitude_hhc::real: sphysfilld(1, fh, sphysreal(wt)); break;
  case amplitude_hhc::sub:  sphysfilld(1, fh, sphyssub(wt));  break;
  default: physfilld(1, fh, wt);
  }
  
  //----- mu = hT/2 -----
  wt = fh*amp(0.25*mu2, 0.25*mu2);
  
  switch(itype) {
  case amplitude_hhc::real: sphysfilld(2, fh, sphysreal(wt)); break;
  case amplitude_hhc::sub:  sphysfilld(2, fh, sphyssub(wt));  break;
  default: physfilld(2, fh, wt);
  }
  
  //----- mu = 2*hT -----
  wt = fh*amp(4.0*mu2, 4.0*mu2);
  
  switch(itype) {
  case amplitude_hhc::real: sphysfilld(3, fh, sphysreal(wt)); break;
  case amplitude_hhc::sub:  sphysfilld(3, fh, sphyssub(wt));  break;
  default: physfilld(3, fh, wt);
  }
}
