//------ DON'T TOUCH THIS PART! ------
#include <bits/hhc-phasespace.h>
#include <bits/hhc-process.h>
#include <bits/hhc-jetfunc.h>
#include <nlo++-module_add.h>

//----- used namespaces -----
using namespace nlo;
using namespace std;


//----- declaration of the user defined functons -----
void inputfunc(unsigned int&, unsigned int&, unsigned int&);
void psinput(phasespace_hhc *, double&);
user_base_hhc * userfunc();

typedef unsigned long int (*module_add_type)(bool, const list<basic_string<char> >&, const basic_string<char>&);
extern  module_add_type module_add;


//----- array of the symbols symbols -----
extern "C"{
  
  struct { 
	const char *name;
	void *address;
  } user_defined_functions[] = 
  {
	//   process index: hhc for hadron-hadron --> jets
	{"procindex", (void *) "hhc"},
  
	//   input function 
	{"inputfunc", (void *) inputfunc},
  
	//   phase space input function 
	{"psinput", (void *) psinput},
  
	//   user defined functions
	{"userfunc",  (void *) userfunc},
  
	//   module to generate the readable result
	{"main_module_add", (void *) module_add},
  
	//  end of the list
	{0, 0}
  };
}
//------ END OF THE DO-NOT-TOUCH-PART ------


//------ USER DEFINED PART STARTS HERE ------
#include "pdf-cteq6.h"


class UserHHC
#if defined(_DENOMINATOR_)
  : public basic_user_set<user0d_hhc, user1h_hhc>
#elif defined(_NUMERATOR_)
  : public user1h_hhc
#endif
{
public:
  //   init and user function
  void initfunc(unsigned int);
  void userfunc(const event_hhc&, const amplitude_hhc&);
  
private:
  //   pdf
  pdf_cteq6 pdf;

  //   private types
  typedef lorentzvector<double> _Lv;
  
  //   particles which contribute to the thrust
  bounded_vector<_Lv> pp;
  
#if defined(_NUMERATOR_)
  //   calculate the thrust
  static void thrust(const bounded_vector<_Lv>&, double&, double&);

  //   sorting the momenta by transverse energy
  struct pT_sort {
    bool operator()(const _Lv& p1, const _Lv& p2) const {
      return p1.perp2() > p2.perp2();
    }
  };  
#endif 
};

//----- defines the module to sum up the results of the different runs -----
module_add_type module_add = 
#if defined(_DENOMINATOR_)
  main_module_add<basic_user_result<user0d_hhc::distbook_type, user1h_hhc::distbook_type> >;
#elif defined(_NUMERATOR_)
  main_module_add<basic_user_result<user1h_hhc::distbook_type> >;
#endif


user_base_hhc * userfunc() {
  return new UserHHC;
}


void inputfunc(unsigned int& nj, unsigned int& nu, unsigned int& nd)
{
  //  number of jets
#if defined(_DENOMINATOR_)
  nj = 2U;
#elif defined(_NUMERATOR_)
  nj = 3U;
#endif

  //  number of the up and down type flavours
  nu = 2U;
  nd = 3U;
} 

void psinput(phasespace_hhc *ps, double& s)
{
  //  total c.m. energy square
  s = 3240000;
  
  //   You can use your own phase generator. 
  //   Here we use the default.
  ps = 0;
} 

void UserHHC::initfunc(unsigned int)
{
#if defined(_DENOMINATOR_)
  user0d_hhc::phys(0, "sigma(hT > 100)");
  user1h_hhc::phys(10, "dsigma/dhT", 45, 50.0, 500.0, 2.0);
#elif defined(_NUMERATOR_)
  phys(1, "tau*dsigma/dtau hT > 100", 25, 0.0, 0.5, 0.2);
  phys(2, "B*dsigma/dB hT > 100", 25, 0.0, 0.5, 0.2);
  
  phys(11, "d<tau>/dhT", 45, 50.0, 500.0, 0.2);
  phys(12, "d<B>/dhT",   45, 50.0, 500.0, 0.2);
#endif
}


#define __fill_hist(id, X, Y)											\
switch(itype) {															\
  case amplitude_hhc::real: sphysfilld(id, X, sphysreal(Y*wt)); break;	\
  case amplitude_hhc::sub:  sphysfilld(id, X, sphyssub(Y*wt));  break;	\
  default: physfilld(id, X, Y*wt);										\
}  


void UserHHC::userfunc(const event_hhc& p, const amplitude_hhc& amp)
{
  double hT = 0.0, emax = 1.0, hTmin = 50.0;
  pp.resize(1,0);
  
  for(int i = 1; i <= p.upper(); i++)
    if(abs(p[i].prapidity()) < emax) {
      hT += p[i].perp();
      pp.push_back(p[i]);
    }
  
  if(pp.size() == 0 || hT < hTmin) return; 
  
  //----- pdf -----
  amplitude_hhc::contrib_type itype = amp.contrib();
  amp.pdf_and_qcd_coupling(pdf, 389385.730);

  //----- hard scale & the weight -----
  double mu2 = hT*hT;
  const weight_hhc& wt = amp(mu2,mu2);
   
#if defined(_DENOMINATOR_)
  if(hT > 100) user0d_hhc::physfill(0, wt);
  __fill_hist(10, hT, 1.0)
#elif defined(_NUMERATOR_)
  //----- fill the weight weights -----
  double T = 0.0, B = 0.0, tau;
  
  std::sort(pp.begin(), pp.end(), pT_sort());
  thrust(pp, T, B); tau = 1.0-T;
  
  if(hT > 100) {
    __fill_hist(1, tau, tau)
    __fill_hist(2, B, B)
  }
  
  __fill_hist(11, hT, tau)
  __fill_hist(12, hT, B)
#endif
}

#if defined(_NUMERATOR_)

void UserHHC::thrust(const bounded_vector<_Lv>& p, double& T, double& B)
{
  int np = p.upper();

  if(np == 1) {
    T = 1.0;
    B = 0.0;
  }
  
  double hT = 0.0;
  threevector<double> s(p[1].X(), p[1].Y(), 0.0);
  
  for(int i = 2; i <= np; i++)
    if(s*p[i] < 0.0) s -= threevector<double>(p[i].X(), p[i].Y(), 0.0);
    else s += threevector<double>(p[i].X(), p[i].Y(), 0.0);
  
  threevector<double> n = s.unit();
  
  T = 0.0; B = 0.0;
  for(int i = 1; i <= np; i++) {
    hT += p[i].perp();
    T += abs(n*p[i]);
    B += cross(n, threevector<double>(p[i].X(), p[i].Y(), 0.0)).mag();
  }
  
  T /= hT; B /= 2.0*hT;
}
#endif
