//------ DON'T TOUCH THIS PART! ------
#include <bits/hhc-phasespace.h>
#include <bits/hhc-process.h>
#include <bits/hhc-jetfunc.h>
#include <nlo++-module_add.h>

//----- used namespaces -----
using namespace nlo;
using namespace std;


//----- declaration of the user defined functons -----
void inputfunc(unsigned int&, unsigned int&, unsigned int&);
void psinput(phasespace_hhc *, double&);
user_base_hhc * userfunc();

typedef unsigned long int (*module_add_type)(bool, const list<basic_string<char> >&, const basic_string<char>&);
extern  module_add_type module_add;


//----- array of the symbols symbols -----
extern "C"{
  
  struct { 
	const char *name;
	void *address;
  } user_defined_functions[] = 
	{
	  //   process index: hhc for hadron-hadron --> jets
	  {"procindex", (void *) "hhc"},
	  
	  //   input function 
	  {"inputfunc", (void *) inputfunc},
	  
	  //   phase space input function 
	  {"psinput", (void *) psinput},
	  
	  //   user defined functions
	  {"userfunc",  (void *) userfunc},
	  
	  //   module to generate the readable result
	  {"main_module_add", (void *) module_add},
	  
	  //  end of the list
	  {0, 0}
	};
}
//------ END OF THE DO-NOT-TOUCH-PART ------

//------ USER DEFINED PART STARTS HERE ------
#include "kT_clus.h"
#include "pdf-cteq6.h"


class UserHHC 
#if defined(_TOTXSEC_)
  : public user1h_hhc
#elif defined(_SUBXSEC_)
  : public user2h_hhc
#endif
{
public:
  //   init and user function
  void initfunc(unsigned int);
  void userfunc(const event_hhc&, const amplitude_hhc&);
  
private:
  //   pdf
  pdf_cteq6 pdf;

  // algorithms
  kT_clus_long jetclus;
  
  // the macro-jet structure
  bounded_vector<lorentzvector<double> > pj; 
};

//----- defines the module to sum up the results of the different runs -----
module_add_type module_add =  
#if defined(_TOTXSEC_)
  main_module_add<basic_user_result<user1h_hhc::distbook_type> >;
#elif defined(_SUBXSEC_)
  main_module_add<basic_user_result<user2h_hhc::distbook_type> >;
#endif

user_base_hhc * userfunc() {
  return new UserHHC;
}

void inputfunc(unsigned int& nj, unsigned int& nu, unsigned int& nd)
{
  //  nuber of jets
#if defined(_TOTXSEC_)
  nj = 2U;
#elif defined(_SUBXSEC_)
  nj = 3U;
#endif

  //  number of the up and down type flavours
  nu = 2U;
  nd = 3U;
} 

void psinput(phasespace_hhc *ps, double& s)
{
  //  total c.m. energy square
  s = 3240000;
  
  //   You can use your own phase generator. 
  //   Here we use the default.
  ps = 0;
} 


void UserHHC::initfunc(unsigned int)
{
#if defined(_TOTXSEC_)
  phys(0, "sigmatot(Ecut)", 20, 10.0, 110.0, 0.01);
#elif defined(_SUBXSEC_)
  phys(3, "sigma3(Ecut, log10(ycut))", 20, 10.0, 110.0, 40, -3.9, 0.1);
  phys(4, "sigma4(Ecut, log10(ycut))", 20, 10.0, 110.0, 40, -3.9, 0.1);
#endif
}


void UserHHC::userfunc(const event_hhc& p, const amplitude_hhc& amp)
{
  typedef lorentzvector<double> _Lv;
  unsigned int np = p.upper();

#if defined(_SUBXSEC_)
  if(amp.integral() == amplitude_hhc::lo) 
    throw "This user function is defined only for NLO calculation.";
#endif
  
  //----- do the cluster analysis----- 
  const bounded_vector<double>& f = jetclus(1.0, p, 1.0);

  //----- jet shape variables -----
  bounded_vector<double> E = f;
  for(unsigned i = np; i > 0; i--) {
    E[i] = sqrt(E[i]);
    E[i] = E[i+1] > E[i] ? E[i+1] : E[i];
  }
  
  //----- Et2 cuts -----
  if(10.0> E[2]) return;
  
  //----- the two macro jets -----
  pj.resize(1,2); jetclus.reco(2, pj);
  
#if defined(_SUBXSEC_)
  //----- resolve the sub-jets -----
  bounded_vector<double> y;
  jetclus.join(E[2], 1.0, y);

  //----- number of subjets -----
  unsigned int ns = 0;
  for(unsigned int i = 1; i <= np; i++)
    if(y[i] > 0.0) ns++;
  
  if(ns == 0) return;
  
  //----- yn jet shapes -----
  double yl = y[2], yh = y[2] > y[1] ? y[2] : y[1];
  if(yl < 0.000125 && yh < 0.000125) return;
#endif  
  
  //----- type of the contribution & pdf -----
  amp.pdf_and_qcd_coupling(pdf, 389385.730);
  
  //----- scales -----
  //double mu = pj[1].perp()+pj[2].perp();
  //double mu2 = mu*mu;
  double mu2 = (pj[1]+pj[2]).mag2();
 
  //----- calculate the weight -----
  const weight_hhc& wt = amp(mu2, mu2);

  //----- fill the weights -----
#if defined(_TOTXSEC_)
  physfillh(0, E[2], wt);
#elif defined(_SUBXSEC_)
  double lyh = log10(yh);
  physfillhh(3, E[2], lyh, wt);
  
  if(amp.contrib() == amplitude_hhc::real && yl > 0.0) {
    double lyl = log10(yl);
    
    physfillhh(3, E[2], lyl, -wt);
    physfillhh(4, E[2], lyl,  wt);
  }
#endif
}
