//------ DON'T TOUCH THIS PART! ------
#include <bits/hhc-phasespace.h>
#include <bits/hhc-process.h>
#include <bits/hhc-jetfunc.h>
#include <nlo++-module_add.h>

//----- used namespaces -----
using namespace nlo;
using namespace std;


//----- declaration of the user defined functons -----
void inputfunc(unsigned int&, unsigned int&, unsigned int&);
void psinput(phasespace_hhc *, double&);
user_base_hhc * userfunc();

typedef unsigned long int (*module_add_type)(bool, const list<basic_string<char> >&, const basic_string<char>&);
extern  module_add_type module_add;


//----- array of the symbols symbols -----
extern "C"{
  
  struct { 
	const char *name;
	void *address;
  } user_defined_functions[] = 
  {
	//   process index: hhc for hadron-hadron --> jets
	{"procindex", (void *) "hhc"},
  
	//   input function 
	{"inputfunc", (void *) inputfunc},
  
	//   phase space input function 
	{"psinput", (void *) psinput},
  
	//   user defined functions
	{"userfunc",  (void *) userfunc},
  
	//   module to generate the readable result
	{"main_module_add", (void *) module_add},
  
	//  end of the list
	{0, 0}
  };
}
//------ END OF THE DO-NOT-TOUCH-PART ------

//------ USER DEFINED PART STARTS HERE ------
#include <algorithm>
#include <kT_clus.h>
#include "pdf-cteq6.h"


class UserHHC 
#if defined(_SCALES_)
  : public user1d_hhc
#elif defined(_XSECTION_)
  : public basic_user_set<user0d_hhc, user1h_hhc, user2h_hhc> 
#endif
{
public:
  //   init and user function
  void initfunc(unsigned int);
  void userfunc(const event_hhc&, const amplitude_hhc&);
  
private:
  //   pdf
  pdf_cteq6 pdf;

  // algorithms
  kT_clus_long jetclus;
   
  //  private types
  typedef lorentzvector<double> _Lv;

  // the jet structore in breit & lab. frame
  bounded_vector<_Lv> cj, pj; 
  bounded_vector<unsigned int> jet;
  
  struct pT_sort {
    bool operator()(const _Lv& p1, const _Lv& p2) const {
      return p1.perp2() > p2.perp2();
    }
  };

  struct E_sort {
    bool operator()(const _Lv& p1, const _Lv& p2) const {
      return p1.T() > p2.T();
    }
  };

#if defined(_SCALES_)
  //  scale studies
  class physscales
  {
  public:
    //   constructor
    physscales() {}
    physscales(const amplitude_hhc& aa, double hsr, double hsf)
      : hard_scale_renorm(hsr), hard_scale_fact(hsf), amp(&aa) {}

    //   set the input parameters
    void operator()(const amplitude_hhc& aa, double hsr, double hsf) {
      amp = &aa; hard_scale_renorm = hsr; hard_scale_fact = hsf;
    }

    //   calculate the weight in the term of scales
    weight_hhc renorm(const double& x) const {
	  return amp -> operator()(x*x*hard_scale_renorm, hard_scale_fact);
    }

    weight_hhc fact(const double& x) const {
      return amp -> operator()(hard_scale_renorm, x*x*hard_scale_fact);
    }

    weight_hhc renorm_and_fact(const double& x) const {
      return amp -> operator()(x*x*hard_scale_renorm, x*x*hard_scale_fact);
    }
    
  private:
    //   private members
    double hard_scale_renorm, hard_scale_fact;
    const amplitude_hhc *amp;
  };
  
  //   helper classes
  physscales scalefunc;
  
  pointer_to_unary_const_member_function<physscales, const double&, weight_hhc>
  renorm, fact, renorm_fact;
  
public:
  //   default constructor
  UserHHC()
    : renorm(scalefunc, &physscales::renorm), fact(scalefunc, &physscales::fact),
      renorm_fact(scalefunc, &physscales::renorm_and_fact) {}  
  
#elif defined(_XSECTION_)
  void fill_jets(int, double, const event_hhc&, const amplitude_hhc&);
#endif
};


//----- defines the module to sum up the results of the different runs -----
module_add_type module_add = 
#if defined(_SCALES_)
main_module_add<basic_user_result<user1d_hhc::distbook_type> >;
#elif defined(_XSECTION_)
main_module_add<basic_user_result<user0d_hhc::distbook_type, user1h_hhc::distbook_type, user2h_hhc::distbook_type> >; 
#endif


user_base_hhc * userfunc() {
  return new UserHHC;
}

void inputfunc(unsigned int& nj, unsigned int& nu, unsigned int& nd)
{
  //  number of jets
  nj = 3U;
  
  //  number of the up and down type flavours
  nu = 2U;
  nd = 3U;
} 

void psinput(phasespace_hhc *ps, double& s)
{
  //  total c.m. energy square
  s = 3240000;
  
  //   You can use your own phase generator. 
  //   Here we use the default.
  ps = 0;
} 



void UserHHC::initfunc(unsigned int)
{
#if defined(_SCALES_)
  phys(1, "kT renorm scale dep.",        60, 0.1, 2.0);
  phys(2, "kT fact scale dep.",          60, 0.1, 2.0);
  phys(3, "kT renorm & fact scale dep.", 60, 0.1, 2.0);
#elif defined(_XSECTION_)
  user1h_hhc::point_type base[] ={
    { 20,  25,  30}, { 30,  35,  40}, { 40,  45,  50}, { 50,  55,  60},
    { 60,  65,  70}, { 70,  75,  80}, { 80,  85,  90}, { 90,  95, 100},
    {100, 105, 110}, {110, 115, 120}, {120, 125, 130}, {130, 135, 140},
    {140, 145, 150}, {150, 155, 160}, {160, 165, 170}, {170, 175, 180},
    {180, 185, 190}, {190, 195, 200}, {200, 205, 210}, {210, 215, 220},
    {220, 225, 230}, {230, 240, 250}, {250, 260, 270}, {270, 280, 290},
    {290, 305, 320}, {320, 335, 350}, {350, 380, 410}, {410, 485, 560}
  };
  
  user0d_hhc::phys(0, "sigma(pT(j) > 20GeV, pT(1)+pT(2)+pT(3) > 175GeV)");
  
  user1h_hhc::phys(1, "pT(1) distribucion xmu = 1.0", 28, base, 0.2);
  user1h_hhc::phys(2, "pT(1) distribucion xmu = 0.5", 28, base, 0.2);
  user1h_hhc::phys(3, "pT(1) distribucion xmu = 2.0", 28, base, 0.2);

  user1h_hhc::phys(4, "X3 Dalitz variable", 40, 0.0, 1.0, 0.2);
  user1h_hhc::phys(5, "X4 Dalitz variable", 40, 0.0, 1.0, 0.2);

  user2h_hhc::phys(0, "dsigma/(dX3dX4) distribution", 20, 0.5, 1.0,  20, 0.5, 1.0);
#endif
}


void UserHHC::userfunc(const event_hhc& p, const amplitude_hhc& amp)
{
  //----- do the cluster analysis-----
  jetclus(p);
  jetclus.incl(cj, jet);

  unsigned int nj = cj.upper();
  if(nj < 3) return;   //  unable to resolve three jets

  //----- cuts -----
  double pT, hT = 0.0;
  double pTmin = 20.0, emax = 2.0;
  
  pj.resize(1,0);
  for(unsigned int i = 1; i <= nj; i++)
    if((pT = cj[i].perp()) > pTmin && abs(cj[i].prapidity()) < emax) {
      hT += pT;
      pj.push_back(cj[i]);
    }
  
  nj = pj.upper();
  if(nj < 3 || hT < 80.0) return;   //  unable to resolve three jets
  
  //----- sorting the jets by transverse energy -----
  std::sort(pj.begin(), pj.end(), pT_sort());
  
  //----- pdf -----
  amp.pdf_and_qcd_coupling(pdf, 389385.730);
 
  //----- hard scale -----
  double mu2 = hT*hT/9.0;

#if defined(_SCALES_)
  //----- set the hard scattering scales -----
  if(hT < 175.0) return;  // no contribution to the total cross section

  scalefunc(amp, mu2, mu2);
  
  //----- calculate the scale dependences -----
  physfill(1, renorm);
  physfill(2, fact);
  physfill(3, renorm_fact);
  
#elif defined(_XSECTION_)
  fill_jets(1, 1.00*mu2, p, amp);
  fill_jets(2, 0.25*mu2, p, amp);
  fill_jets(3, 2.00*mu2, p, amp);

  if(hT < 175.0) return;  // no contribution to the Dalitz variables
  
  //----- calculate & fill the weight -----
  amplitude_hhc::contrib_type itype = amp.contrib();
  const weight_hhc& wt = amp(mu2,mu2);

  //----- total 3-jet cross section -----
  user0d_hhc::physfill(0, wt);

  //----- Dalitz variables -----
  lorentzvector<double> q = pj[1]+pj[2]+pj[3];
  threevector<double> bVec = -q.boostVector();
  pj[1].boost(bVec); pj[2].boost(bVec); pj[3].boost(bVec);
  
  //----- sorting the three leading jets by energy -----
  std::sort(pj.begin(), pj.begin()+3, E_sort());
  
  double MJ = q.mag();
  double X3 = 2.0*pj[1].T()/MJ, X4 = 2.0*pj[2].T()/MJ;
  
  //----- fill the weights -----  
  switch(itype) {
  case amplitude_hhc::real: sphysfilld(4, X3, sphysreal(wt)); break;
  case amplitude_hhc::sub:  sphysfilld(4, X3, sphyssub(wt));  break;
  default: physfilld(4, X3, wt);
  }
  
  switch(itype) {
  case amplitude_hhc::real: sphysfilld(5, X4, sphysreal(wt)); break;
  case amplitude_hhc::sub:  sphysfilld(5, X4, sphyssub(wt));  break;
  default: physfilld(5, X4, wt);
  }
  
  physfilldd(0, X3, X4, wt);
#endif
}


#if defined(_XSECTION_)
void UserHHC::fill_jets(int id, double mu2, const event_hhc& p, const amplitude_hhc& amp)
{ 
  //----- hard scale -----
  double pT = pj[1].perp();
  
  //----- calculate & fill the weight -----
  const weight_hhc& wt = amp(mu2,mu2);
  
  if(pT < 250)
    switch(amp.contrib()) {
    case amplitude_hhc::real: sphysfilld(id, pT, sphysreal(wt)); break;
    case amplitude_hhc::sub:  sphysfilld(id, pT, sphyssub(wt));  break;
    default: physfilld(id, pT, wt);
    }
  else physfilld(id, pT, wt);
}
#endif
